from .deezer import *
from .soundcloud import *
from .youtube import *

import logging, shutil, os, atexit
from sys import exit
__name__ == 'music-helper'
__version__ = '2.0.0'


class DummyClass:    
    def __getattr__(self, name):
        def dummy_method(*args, **kwargs):
            raise ValueError('This provider has not been configured.')
        return dummy_method


class MusicHelper(object):
    def __init__(self, deezer_arl:str=None,sc_data:tuple=None, ytm_oauth:str="", debug:bool=False, ffmpeg_path:str="./") -> None:
        atexit.register(self.save)
        
        self.deezer_status = False
        self.ytm_status = False
        self.yt_status = True
        self.sc_status = False
        self.debug = debug
        self.root_path = os.path.dirname(os.path.abspath(__file__))
        if ffmpeg_path:
            self.ffmpeg_path = shutil.which(ffmpeg_path)
        else:
            self.ffmpeg_path = shutil.which('ffmpeg')

        if self.ffmpeg_path is None:
            if debug: logging.critical('Could not find ffmpeg.' )
            exit('Could not find ffmpeg' if not debug else 1)
       


        #### INIT DEEZER
        if deezer_arl:
            self.deezer = Deezer(deezer_arl, debug)
            self.deezer_status = True
            if debug: logging.info("Deezer status: Running")
        else: self.deezer = DummyClass()

        #### INIT YOUTUBE
        self.youtube = YouTube(debug, self.ffmpeg_path)
        self.yt_status = True
        if debug: logging.info("YouTube status: Running")

        #### INIT YOUTUBE MUSIC
        if ytm_oauth:
            self.ytm = YouTubeMusic(oauth = ytm_oauth, debug = self.debug)
            self.ytm_status = True
            if debug: logging.info("YouTube Music status: Running")
        else: self.ytm = DummyClass()

        #### INIT SoundCloud
        if sc_data:
            self.soundcloud = SoundCloud(client_id=sc_data[0],auth_token=sc_data[1], 
                                         debug = self.debug)
            self.sc_status = True
            if debug: logging.info("SoundCloud status: Running")
        else: self.soundcloud = DummyClass()



    def save(self):
        ...


from youtubesearchpython.__future__ import VideosSearch
import io, asyncio, os
from ..util import sanitize_string, apply_tags_to_audio
from ..deezer.util import create_folders
from pytube import YouTube as pYouTube
import ytmusicapi, os, logging, asyncio


FFMPEG_PATH = ""
class YouTube:
    def __init__(self, debug:bool, ffmpeg:str) -> None:
        global FFMPEG_PATH
        self.debug = debug
        FFMPEG_PATH = ffmpeg

    async def search(self, query:str, limit:int=10, clear_result:bool=False):
        search_result = VideosSearch(query, limit = limit)
        search_result = await search_result.next()
        if clear_result:
            temp_list = []
            for item in search_result['result']:
                time_parts = item.get('duration', "1:10").split(':')
                if len(time_parts) == 2: 
                    minutes, seconds = map(int, time_parts)
                    total_seconds = (minutes * 60) + seconds
                elif len(time_parts) == 3:  
                    hours, minutes, seconds = map(int, time_parts)
                    total_seconds = (hours * 3600) + (minutes * 60) + seconds
                
                if total_seconds < 360:
                    video_title = item.get('title', "unknown")
                    video_user = item.get('channel', {}).get('name', "unknown")

                    video_title, video_channelname = await sanitize_string(video_title, video_user)
                    temp_list.append({
                        'trackID': item['id'],
                        'title': video_title,
                        'artist': video_channelname,
                        'album': video_title,
                        'cover': item.get('thumbnails', [])[0].get('url', '')
                    })
            
            search_result['result'] = temp_list

        return search_result
    
    @staticmethod
    async def download_audio(video_id:str, download_dir:str=None, tags:dict=None):

        a = pYouTube(f"https://youtu.be/{video_id}", 
                        use_oauth=True,
                        allow_oauth_cache=True)
        if not tags:
            tags = {
                'artist': a.author,
                'title': a.title,
                'album': a.title,
                'cover': a.thumbnail_url
            }
        if download_dir:
            create_folders(download_dir)
            artist, title = await sanitize_string(tags.get('artist', ''), tags.get('title'), True)
            filename = f"{artist} - {title}.mp3"
            filepath = os.path.join(download_dir, filename)

        audio_stream = a.streams.filter(only_audio=True, file_extension='mp4').first()
        audio_length = a.length
        tags.update({'cover': a.thumbnail_url})

        start = 0
        end = audio_length - 1
        cmd = [
            FFMPEG_PATH, "-ss", str(start), "-i", audio_stream.url,
            "-t", str(end - start + 1), "-acodec", "libmp3lame",
            "-f", "mp3", filepath if download_dir else '-'
        ]
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        mp3_data, _ = await proc.communicate()
        if proc.returncode == 0:
            if not download_dir:
                audio_io_data = io.BytesIO(mp3_data)

                audio_io_data = await apply_tags_to_audio(
                    audio = audio_io_data, 
                    tags = tags, 
                    return_bytes = True)
                
                return audio_io_data
            else:
                await apply_tags_to_audio(
                    audio = filepath, 
                    tags = tags, 
                    return_bytes = False)
                
                return filepath





class YouTubeMusic:
    def __init__(self, oauth:str, debug:bool) -> None:
        self.debug = debug
        self.oauth = oauth
        try:
            if os.path.exists(self.oauth):
                self.ytm = ytmusicapi.YTMusic(self.oauth)
            else:
                ytmusicapi.setup_oauth(self.oauth)
                self.ytm = ytmusicapi.YTMusic(self.oauth)
        except Exception as ytm_exception:
            raise ytm_exception


    async def __search(self, query:str, filter:str, scope:str, 
                       limit:int, ignore_spelling:bool):
        data = self.ytm.search(query = query, filter = filter, scope = scope, 
                               limit = limit, ignore_spelling = ignore_spelling)
        return data 
    async def search(self, query:str, filter:str=None, scope:str=None, 
                       limit:int=20, ignore_spelling:bool=False):
        """
        Search YouTube music
        Returns results within the provided category.

        :param query: Query string, i.e. 'Oasis Wonderwall'
        :param filter: Filter for item types. Allowed values: ``songs``, ``videos``, ``albums``, ``artists``, ``playlists``, ``community_playlists``, ``featured_playlists``, ``uploads``.
          Default: Default search, including all types of items.
        :param scope: Search scope. Allowed values: ``library``, ``uploads``.
            For uploads, no filter can be set! An exception will be thrown if you attempt to do so.
            Default: Search the public YouTube Music catalogue.
        :param limit: Number of search results to return
          Default: 20
        :param ignore_spelling: Whether to ignore YTM spelling suggestions.
          If True, the exact search term will be searched for, and will not be corrected.
          This does not have any effect when the filter is set to ``uploads``.
          Default: False, will use YTM's default behavior of autocorrecting the search.
        :return: List of results depending on filter.
          resultType specifies the type of item (important for default search).
          albums, artists and playlists additionally contain a browseId, corresponding to
          albumId, channelId and playlistId (browseId=``VL``+playlistId)

          Example list for default search with one result per resultType for brevity. Normally
          there are 3 results per resultType and an additional ``thumbnails`` key::

            [
              {
                "category": "Top result",
                "resultType": "video",
                "videoId": "vU05Eksc_iM",
                "title": "Wonderwall",
                "artists": [
                  {
                    "name": "Oasis",
                    "id": "UCmMUZbaYdNH0bEd1PAlAqsA"
                  }
                ],
                "views": "1.4M",
                "videoType": "MUSIC_VIDEO_TYPE_OMV",
                "duration": "4:38",
                "duration_seconds": 278
              },
              {
                "category": "Songs",
                "resultType": "song",
                "videoId": "ZrOKjDZOtkA",
                "title": "Wonderwall",
                "artists": [
                  {
                    "name": "Oasis",
                    "id": "UCmMUZbaYdNH0bEd1PAlAqsA"
                  }
                ],
                "album": {
                  "name": "(What's The Story) Morning Glory? (Remastered)",
                  "id": "MPREb_9nqEki4ZDpp"
                },
                "duration": "4:19",
                "duration_seconds": 259
                "isExplicit": false,
                "feedbackTokens": {
                  "add": null,
                  "remove": null
                }
              },
              {
                "category": "Albums",
                "resultType": "album",
                "browseId": "MPREb_9nqEki4ZDpp",
                "title": "(What's The Story) Morning Glory? (Remastered)",
                "type": "Album",
                "artist": "Oasis",
                "year": "1995",
                "isExplicit": false
              },
              {
                "category": "Community playlists",
                "resultType": "playlist",
                "browseId": "VLPLK1PkWQlWtnNfovRdGWpKffO1Wdi2kvDx",
                "title": "Wonderwall - Oasis",
                "author": "Tate Henderson",
                "itemCount": "174"
              },
              {
                "category": "Videos",
                "resultType": "video",
                "videoId": "bx1Bh8ZvH84",
                "title": "Wonderwall",
                "artists": [
                  {
                    "name": "Oasis",
                    "id": "UCmMUZbaYdNH0bEd1PAlAqsA"
                  }
                ],
                "views": "386M",
                "duration": "4:38",
                "duration_seconds": 278
              },
              {
                "category": "Artists",
                "resultType": "artist",
                "browseId": "UCmMUZbaYdNH0bEd1PAlAqsA",
                "artist": "Oasis",
                "shuffleId": "RDAOkjHYJjL1a3xspEyVkhHAsg",
                "radioId": "RDEMkjHYJjL1a3xspEyVkhHAsg"
              }
            ]


        """
        task = await asyncio.get_event_loop().run_in_executor(None, 
                        lambda: self.__search(query = query, filter = filter, 
                                             scope = scope, limit = limit, 
                                             ignore_spelling = ignore_spelling))
        
        
        task_result = await task
        return task_result 
    
    
    async def __search_by_id(self, videoID:str, signatureTimestamp:int):
        data = self.ytm.get_song(videoId=videoID, signatureTimestamp=signatureTimestamp)
        return data
    async def search_by_id(self, videoID:str, signatureTimestamp:int = None):
        task = await asyncio.get_event_loop().run_in_executor(None, 
                        lambda: self.__search_by_id(videoID = videoID, signatureTimestamp = signatureTimestamp))
        
        
        task_result = await task
        return task_result
    



import asyncio
from .constants import *
from .exceptions import *
from .util import *
from ..util import apply_tags_to_audio

from typing import Union
import httpx, logging, tempfile, io, aiofiles

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes



httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.ERROR)
class Deezer:
    def __init__(self, arl, debug) -> None:
        self.arl = arl
        self.cookies = {'arl': self.arl}
        self.debug = debug
            
        self.token = None
        self.user = {}
        self.__temp_cookies = {}

        asyncio.get_event_loop().run_until_complete(self.get_user_data())
        self.cookies = self.__temp_cookies

    ###
    ### API methods
    ###
    async def get_cookies(self):
        return self.cookies if self.cookies else None
    async def _api_call(self, method, params={}):
        token = self.token if self.token else 'null'
        if method != api_methods.GET_USER_DATA:
            token = self.token


        async with httpx.AsyncClient(http2=True,
            headers=networking_settings.HTTP_HEADERS, params={
            "api_version": "1.0",
            "api_token": token,
            "input": "3",
            "method": method
        }, cookies=await self.get_cookies()) as client:
            response = await client.post(api_urls.API_URL, json=params)
            data = response.json()
            self.__temp_cookies = response.cookies
                 




        if "error" in data and data["error"]:
            error_type = list(data["error"].keys())[0]
            error_message = data["error"][error_type]

            raise APIRequestError(
                "{0} : {1}".format(error_type, error_message))

        return data
    async def _legacy_search(self, method, query, limit=30, index=0):
        query = clean_query(query)

        data = await self._legacy_api_call(method, {
            "q": query,
            "limit": limit,
            "index": index
        })

        return data["data"]
    async def _legacy_api_call(self, method, params={}):
        url = "{0}/{1}".format(api_urls.LEGACY_API_URL, method)

        async with httpx.AsyncClient(headers=networking_settings.HTTP_HEADERS, params=params, cookies=await self.get_cookies()) as client:
            response = await client.get(url)
            data = response.json()

            
        if "error" in data and data["error"]:
            error_type = list(data["error"].keys())[0]
            error_message = data["error"][error_type]
            raise APIRequestError(
                "{0} : {1}".format(error_type, error_message))

        return data  
    async def get_user_data(self):
        """Gets the data of the user, this will only work arl is the cookie. Make sure you have run login_via_arl() before using this.

        Raises:
            LoginError: Will raise if the arl given is not identified by Deezer
        """

        data = await self._api_call(api_methods.GET_USER_DATA)


        data = data['results']
        self.token = data["checkForm"]

        if not data["USER"]["USER_ID"]:
            raise LoginError("Arl is invalid.")

        raw_user = data["USER"]


        if raw_user["USER_PICTURE"]:
            self.user = {
                "id": raw_user["USER_ID"],
                "name": raw_user["BLOG_NAME"],
                "arl": self.cookies["arl"],
                "image": "https://e-cdns-images.dzcdn.net/images/user/{0}/250x250-000000-80-0-0.jpg".format(raw_user["USER_PICTURE"])
            }
        else:
            self.user = {
                "id": raw_user["USER_ID"],
                "name": raw_user["BLOG_NAME"],
                "arl": self.cookies["arl"],
                "image": "https://e-cdns-images.dzcdn.net/images/user/250x250-000000-80-0-0.jpg"
            }


    """
        ALBUM
    """
    async def get_album(self, album_id):
        """Gets the album data of the given {album_id}

        Arguments:
            album_id {str} -- Album Id

        Returns:
            dict -- Album data
        """

        data = await self._legacy_api_call("/album/{0}".format(album_id))
        if data["cover_small"]:
            data["cover_id"] = str(data["cover_small"]).split(
                "cover/")[1].split("/")[0]
        else:
            data["cover_id"] = -1

        return data
    async def get_album_poster(self, album, size=500, ext="jpg"):
        """Gets the album poster as a binary data

        Arguments:
            album {dict} -- Album data

        Keyword Arguments:
            size {int} -- Size of the image, {size}x{size} (default: {500})
            ext {str} -- Extension of the image, can be ('.jpg' or '.png') (default: {"jpg"})

        Returns:
            bytes -- Binary data of the image
        """
        return await self._get_poster(album["cover_id"], size=size, ext=ext)
    async def _get_poster(self, poster_id, size=500, ext="jpg"):
        ext = ext.lower()
        if ext != "jpg" and ext != "png":
            raise ValueError("Image extension should only be jpg or png!")
        if poster_id == -1:
            url = "https://static.vecteezy.com/system/resources/thumbnails/022/059/000/small/no-image-available-icon-vector.jpg"
        else:
            url = f'https://e-cdns-images.dzcdn.net/images/cover/{poster_id}/{size}x{size}.{ext}'
        

        async with httpx.AsyncClient(headers=networking_settings.HTTP_HEADERS, cookies=await self.get_cookies()) as client:
            res = await client.get(url)
            image_bytes = res.read()

                
        return {
            "image": image_bytes,
            "size": (size, size),
            "ext": ext,
            "mime_type": "image/jpeg" if ext == "jpg" else "image/png"
        }
    async def get_album_tracks(self, album_id):
        """Gets the tracks of the given {album_id}

        Arguments:
            album_id {str} -- Album Id

        Returns:
            list -- List of tracks
        """

        data = await self._api_call(api_methods.ALBUM_TRACKS, params={
            "ALB_ID": album_id,
            "NB": -1
        })

        for i, track in enumerate(data["results"]["data"]):
            track["_POSITION"] = i + 1

        return data["results"]["data"]
    async def search_albums(self, query, limit=30, index=0):
        """Searches albums on a given query

        Arguments:
            query {str} -- Query keyword

        Keyword Arguments:
            limit {int} -- Number of results (default: {30})
            index {int} -- Offset (default: {0})

        Returns:
            list -- List of albums
        """

        return await self._legacy_search(api_methods.SEARCH_ALBUM, query, limit=limit, index=index)
    """
        ARTIST
    """
    async def get_artist(self, artist_id):
        """Gets the artist data from the given {artist_id}

        Arguments:
            artist_id {str} -- Artist Id

        Returns:
            dict -- Artist data
        """

        data = await self._api_call(api_methods.PAGE_ARTIST, params={
            "ART_ID": artist_id,
            "LANG": "en"
        })

        return data["results"]
    async def get_artist_poster(self, artist, size=500, ext="jpg"):
        """Gets the artist poster as a binary data

        Arguments:
            artist {dict} -- artist data

        Keyword Arguments:
            size {int} -- Size of the image, {size}x{size} (default: {500})
            ext {str} -- Extension of the image, can be ('.jpg' or '.png') (default: {"jpg"})

        Returns:
            bytes -- Binary data of the image
        """

        if not "ART_PICTURE" in artist and "DATA" in artist:
            artist = artist["DATA"]

        return await self._get_poster(artist["ART_PICTURE"], size=size, ext=ext)
    async def get_artist_discography(self, artist_id):
        """Gets the artist's discography (tracks)

        Arguments:
            artist_id {str} -- Artist Id

        Returns:
            dict -- Artist discography data
        """

        data = await self._api_call(api_methods.ARTIST_DISCOGRAPHY, params={
            "ART_ID": artist_id,
            "NB": 500,
            "NB_SONGS": -1,
            "START": 0
        })

        return data["results"]["data"]
    async def get_artist_top_tracks(self, artist_id):
        """Gets the top tracks of the given artist

        Arguments:
            artist_id {str} -- Artist Id

        Returns:
            list -- List of track
        """

        data = await self._api_call(api_methods.ARTIST_TOP_TRACKS, params={
            "ART_ID": artist_id,
            "NB": 100
        })

        for i, track in enumerate(data["results"]["data"]):
            track["_POSITION"] = i + 1

        return data["results"]["data"]
    async def search_artists(self, query, limit=30, index=0):
        """Searches artists on a given query

        Arguments:
            query {str} -- Query keyword

        Keyword Arguments:
            limit {int} -- Number of tracks (default: {30})
            index {int} -- Offset (default: {0})

        Returns:
            list -- List of artists
        """

        return await self._legacy_search(api_methods.SEARCH_ARTIST, query, limit=limit, index=index)
    """
        PLAYLIST
    """
    async def get_playlist(self, playlist_id):
        """Gets the playlist data from the given playlist_id

        Arguments:
            playlist_id {str} -- Playlist Id

        Returns:
            dict -- Playlist data
        """

        data = await self._api_call(api_methods.PAGE_PLAYLIST, params={
            "playlist_id": playlist_id,
            "LANG": "en"
        })

        return data["results"]
    async def get_playlist_tracks(self, playlist_id):
        """Gets the tracks inside the playlist

        Arguments:
            playlist_id {str} -- Playlist Id

        Returns:
            list -- List of tracks
        """

        data = await self._api_call(api_methods.PLAYLIST_TRACKS, params={
            "PLAYLIST_ID": playlist_id,
            "NB": -1
        })

        for i, track in enumerate(data["results"]["data"]):
            track["_POSITION"] = i + 1

        return data["results"]["data"]
    async def search_playlists(self, query, limit=30, index=0):
        """Searches playlists on a given query

        Arguments:
            query {str} -- Query keyword

        Keyword Arguments:
            limit {int} -- Number of tracks (default: {30})
            index {int} -- Offset (default: {0})

        Returns:
            list -- List of playlists
        """

        return await self._legacy_search(api_methods.SEARCH_PLAYLIST, query, limit=limit, index=index)

    """
        TRACKS
    """
    async def search_tracks(self, query, limit=30, index=0):
        """Searches tracks on a given query

        Arguments:
            query {str} -- Query keyword

        Keyword Arguments:
            limit {int} -- Number of results (default: {30})
            index {int} -- Offset (default: {0})

        Returns:
            list -- List of tracks
        """

        return await self._legacy_search(api_methods.SEARCH_TRACK, query, limit=limit, index=index)
    async def get_track_tags(self, track, separator=", ", with_cover:bool=True):
        """Gets the possible ID3 tags of the track.

        Arguments:
            track {dict} -- Track dictionary, similar to the {info} value that is returned {using get_track()}

        Keyword Arguments:
            separator {str} -- Separator to separate multiple artists (default: {", "})

        Returns:
            dict -- Tags
        """

        track = track["DATA"] if "DATA" in track else track

        album_data = await self.get_album(track["ALB_ID"])

        if "main_artist" in track["SNG_CONTRIBUTORS"]:
            main_artists = track["SNG_CONTRIBUTORS"]["main_artist"]
            artists = main_artists[0]
            for i in range(1, len(main_artists)):
                artists += separator + main_artists[i]
        else:
            artists = track["ART_NAME"]

        title = track["SNG_TITLE"]

        if "VERSION" in track and track["VERSION"] != "":
            title += " " + track["VERSION"]

        def should_include_featuring():
            feat_keywords = ["feat.", "featuring", "ft."]

            for keyword in feat_keywords:
                if keyword in title.lower():
                    return False
            return True

        if should_include_featuring() and "featuring" in track["SNG_CONTRIBUTORS"]:
            featuring_artists_data = track["SNG_CONTRIBUTORS"]["featuring"]
            featuring_artists = featuring_artists_data[0]
            for i in range(1, len(featuring_artists_data)):
                featuring_artists += separator + featuring_artists_data[i]

            title += f" (feat. {featuring_artists})"

        total_tracks = album_data["nb_tracks"]
        track_number = str(track["TRACK_NUMBER"]) + "/" + str(total_tracks)

        if with_cover:
            cover = await self.get_album_poster(album_data, size=1000)

        tags = {
            "title": title,
            "artist": artists,
            "genre": None,
            "album": track["ALB_TITLE"],
            "albumartist": track["ART_NAME"],
            "label": album_data["label"],
            "date": track["PHYSICAL_RELEASE_DATE"],
            "discnumber": track["DISK_NUMBER"],
            "tracknumber": track_number,
            "isrc": track["ISRC"],
            "copyright": track["COPYRIGHT"],
            "_albumart": cover if with_cover else None,
        }

        if len(album_data["genres"]["data"]) > 0:
            tags["genre"] = album_data["genres"]["data"][0]["name"]

        if "author" in track["SNG_CONTRIBUTORS"]:
            _authors = track["SNG_CONTRIBUTORS"]["author"]

            authors = _authors[0]
            for i in range(1, len(_authors)):
                authors += separator + _authors[i]

            tags["author"] = authors

        return tags
    async def get_track(self, track_id, with_cover:bool=True):
        """Gets the track info using the Deezer API

        Arguments:
            track_id {str} -- Track Id

        Returns:
            dict -- Dictionary that contains the {info}, {download} partial function, {tags}, and {get_tag} partial function.
        """

        method = api_methods.SONG_GET_DATA
        params = {
            "SNG_ID": track_id
        }

        if not int(track_id) < 0:
            method = api_methods.PAGE_TRACK

        data = await self._api_call(method, params=params)
        data = data["results"]

        return {
            "info": data,
            "tags": await self.get_track_tags(data, with_cover=with_cover),
        }
    async def get_track_download_url(self, track, quality=None, fallback=True, renew=False, **kwargs):
        """Gets and decrypts the download url of the given track in the given quality

        Arguments:
            track {dict} -- Track dictionary, similar to the {info} value that is returned {using get_track()}

        Keyword Arguments:
            quality {str} -- Use values from {track_formats}, will get the default quality if None or an invalid is given. (default: {None})
            fallback {bool} -- Set to True to if you want to use fallback qualities when the given quality is not available. (default: {False})
            renew {bool} -- Will renew the track object (default: {False})

        Raises:
            DownloadLinkDecryptionError: Will be raised if the track dictionary does not have an MD5
            ValueError: Will be raised if valid track argument was given

        Returns:
            str -- Download url
        """

        if renew:
            track = await self.get_track(track["SNG_ID"])
            track = track["info"]

        if not quality:
            quality = track_formats.MP3_128
            fallback = True

        try:
            track = track["DATA"] if "DATA" in track else track

            if not "MD5_ORIGIN" in track:
                raise DownloadLinkDecryptionError(
                    "MD5 is needed to decrypt the download link.")

            md5_origin = track["MD5_ORIGIN"]
            track_id = track["SNG_ID"]
            media_version = track["MEDIA_VERSION"]
        except ValueError:
            raise ValueError(
                "You have passed an invalid argument. This method needs the \"DATA\" value in the dictionary returned by the get_track() method.")

        def decrypt_url(quality_code):
            magic_char = "¤"
            step1 = magic_char.join((md5_origin,
                                     str(quality_code),
                                     track_id,
                                     media_version))
            m = hashlib.md5()
            m.update(bytes([ord(x) for x in step1]))

            step2 = m.hexdigest() + magic_char + step1 + magic_char
            step2 = step2.ljust(80, " ")

            cipher = Cipher(algorithms.AES(bytes('jo6aey6haid2Teih', 'ascii')),
                            modes.ECB(), default_backend())

            encryptor = cipher.encryptor()
            step3 = encryptor.update(bytes([ord(x) for x in step2])).hex()

            cdn = track["MD5_ORIGIN"][0]

            return f'https://e-cdns-proxy-{cdn}.dzcdn.net/mobile/1/{step3}'

        url = decrypt_url(track_formats.TRACK_FORMAT_MAP[quality]["code"])


        async with httpx.AsyncClient(headers=networking_settings.HTTP_HEADERS, cookies=await self.get_cookies()) as client:
            response = await client.get(url)
            if not fallback or (response.status_code == 200 and int(response.headers.get("Content-Length", 0)) > 0):
                return (url, quality)
            else:
                if "fallback_qualities" in kwargs:
                    fallback_qualities = kwargs["fallback_qualities"]
                else:
                    fallback_qualities = track_formats.FALLBACK_QUALITIES

                for key in fallback_qualities:
                    url = decrypt_url(
                        track_formats.TRACK_FORMAT_MAP[key]["code"])

                    async with httpx.AsyncClient(headers=networking_settings.HTTP_HEADERS, cookies=await self.get_cookies()) as client:
                        response = await client.get(url)
                        if not fallback or (response.status == 200 and int(response.headers.get("Content-Length", 0)) > 0):
                            return (url, key) 
    async def download_track(self, track, download_dir=None, quality=None, fallback=True, filename=None, 
                             renew=False,with_metadata=True, with_lyrics=True, 
                             tag_separator=", ", **kwargs):
        
        if with_lyrics:
            if "LYRICS" in track:
                lyric_data = track["LYRICS"]
            else:
                try:
                    if "DATA" in track:
                        _lyric_data = await self.get_track_lyrics(
                            track["DATA"]["SNG_ID"])
                        lyric_data = _lyric_data["info"]
                    else:
                        _lyric_data = await self.get_track_lyrics(
                            track["SNG_ID"])
                        lyric_data = _lyric_data["info"]
                except APIRequestError:
                    with_lyrics = False

        track = track["DATA"] if "DATA" in track else track
        tags = await self.get_track_tags(track, separator=tag_separator, with_cover=with_metadata)
        title = tags["title"]
        artist = tags['artist']
        if self.debug: logging.info(f'[DEEZER]: Download started [{artist} - {title}]')
        url, quality_key = await self.get_track_download_url(
            track, quality, fallback=fallback, renew=renew, **kwargs)
        blowfish_key = get_blowfish_key(track["SNG_ID"])

        quality = track_formats.TRACK_FORMAT_MAP[quality_key]
        ext = quality["ext"]

        if not filename:
            filename =  f"{artist} - {title}{ext}"
        if not str(filename).endswith(ext):
            filename += ext

        filename = clean_filename(filename)
        if download_dir:
            download_dir = path.normpath(download_dir)
            download_path = path.join(download_dir, filename)
            create_folders(download_dir)

        chunk_size = 2 * 1024
        async with httpx.AsyncClient(headers=networking_settings.HTTP_HEADERS, cookies=self.cookies) as client:
            res = await client.get(url, follow_redirects=True)
            data_iter = res.iter_bytes(chunk_size)

            i = 0

            if download_dir:
                async with aiofiles.open(download_path, 'wb') as f:
                    f.seek(0)
                
                    for chunk in data_iter:
                        try:
                            if i % 3 > 0:
                                await f.write(chunk)
                            elif len(chunk) < chunk_size:
                                await f.write(chunk)
                                break
                            else:
                                cipher = Cipher(algorithms.Blowfish(blowfish_key),
                                            modes.CBC(
                                                bytes([i for i in range(8)])),
                                            default_backend())

                                decryptor = cipher.decryptor()
                                dec_data = decryptor.update(
                                    chunk) + decryptor.finalize()

                                await f.write(dec_data)
                        except:
                            if self.debug: logging.error(f"[DEEZER] DL-ERROR: (bad chunk) - {type(chunk)}")
                        i += 1
            else:
                audio = io.BytesIO()
                audio.seek(0)
                for chunk in data_iter:
                    try:
                        if i % 3 > 0 or len(chunk) < chunk_size:
                            audio.write(chunk)
                        else:
                            cipher = Cipher(algorithms.Blowfish(blowfish_key),
                                            modes.CBC(bytes([i for i in range(8)])),
                                            default_backend())
                            decryptor = cipher.decryptor()
                            dec_data = decryptor.update(chunk) + decryptor.finalize()
                            audio.write(dec_data)

                    except Exception as e:
                        if self.debug: logging.error(f"[DEEZER] DL-ERROR: (bad chunk) - {type(chunk)}\nERROR: {e}")
                    
                    i += 1
                audio.seek(0)

        if with_metadata:
            if self.debug: logging.info(f'[DEEZER]: Tag editing [{artist} - {title}]')
            if ext.lower() == ".flac":
                pass #self._write_flac_tags(download_path, track, tags=tags)
            else:
                if download_dir:
                    await self._write_mp3_tags(audio=download_path, 
                                               track=track, tags=tags)
                else:
                    audio = await self._write_mp3_tags(audio=audio, 
                                               track=track, tags=tags, return_bytes=True)
                
        if with_lyrics:
            lyrics_path = tempfile.mkstemp(prefix=filename[:-len(ext)], 
                                           suffix='.lrc')
            lyric_check = await self.save_lyrics(lyric_data, lyrics_path[1])



        return {
            'track': download_path if download_dir else audio,
            'lyric': lyric_check[1] if with_lyrics and lyric_check[0] else None
        }


    """
        TAGS
    """
    async def _write_mp3_tags(self, audio:Union[str, io.BytesIO], track, tags=None):
        track = track["DATA"] if "DATA" in track else track

        if not tags:
            tags = await self.get_track_tags(track)

        _task = await asyncio.get_event_loop().run_in_executor(None, 
                                    lambda: apply_tags_to_audio(audio, tags))
        result = await _task
        return result

    """
        LYRIC
    """
    async def get_track_lyrics(self, track_id):
        """Gets the lyrics data of the given {track_id}

        Arguments:
            track_id {str} -- Track Id

        Returns:
            dict -- Dictionary that containts the {info}, and {save} partial function.
        """

        data = await self._api_call(api_methods.SONG_LYRICS, params={
            "SNG_ID": track_id
        })
        data = data["results"]

        return {
            "info": data
        }
    async def save_lyrics(self, lyric_data, save_path):
        """Saves the {lyric_data} into a .lrc file.

        Arguments:
            lyric_data {dict} -- The 'info' value returned from {get_track_lyrics()}
            save_path {str} -- Full path on where the file is to be saved

        Returns:
            bool -- Operation success
        """

        filename = path.basename(save_path)
        filename = clean_filename(filename)
        save_path = path.join(path.dirname(save_path), filename)

        if not str(save_path).endswith(".lrc"):
            save_path += ".lrc"

        create_folders(path.dirname(save_path))

        async with aiofiles.open(save_path, 'w', encoding="utf-8") as f:
            if not "LYRICS_SYNC_JSON" in lyric_data:
                return (False, save_path)

            sync_data = lyric_data["LYRICS_SYNC_JSON"]

            for line in sync_data:
                if str(line["line"]):
                    await f.write("{0}{1}".format(
                        line["lrc_timestamp"], line["line"]))
                await f.write("\n")

        return (True, save_path)



from mutagen.mp3 import MP3, HeaderNotFoundError
from mutagen.id3 import ID3, APIC, TIT2, TPE1, TALB
from typing import Union
import io, aiohttp, unidecode, re
from .deezer.constants.networking_settings import HTTP_HEADERS

SANITIZE_PARTS = ["lyric","lyrics", "(music video)","(official music video)", "feat.", "f."]

@staticmethod
async def sanitize_string(string_line:str, other_string_line:str="", _unidecode:bool=False):
    if _unidecode: string_line = unidecode.unidecode_expect_nonascii(string_line) 
    string_line = re.sub(r'[^\w\s.-]', '', string_line)

    if other_string_line:
        string_line = string_line.replace(other_string_line, "")
        if _unidecode:other_string_line = unidecode.unidecode_expect_nonascii(other_string_line) 
        other_string_line = re.sub(r'[^\w\s.-]', '', other_string_line)

    string_line = string_line.replace("-", "")

    for part in SANITIZE_PARTS:
        if string_line.lower().find(part) != -1:
            string_line = string_line.replace(part, "")

        if other_string_line:
            if other_string_line.lower().find(part) != -1:
                other_string_line = other_string_line.replace(part, "")

    

    if other_string_line:
        return (string_line.strip(), other_string_line.strip())

    
    return string_line.strip()  
@staticmethod
async def apply_tags_to_audio(audio:Union[str, io.BytesIO], tags:dict) -> Union[str, io.BytesIO]: 
    if tags:
        try:
            audio_file = MP3(audio, ID3=ID3)
        except HeaderNotFoundError:
            audio = io.BytesIO(audio.getvalue())
            try:
                audio_file = MP3(audio, ID3=ID3)
            except HeaderNotFoundError:
                return None
    if not audio_file.tags:
            audio_file.tags = ID3()

    audio_file.tags.add(TPE1(encoding=3, text=tags.get('artist', "unknown")))  
    audio_file.tags.add(TALB(encoding=3, text=tags.get('album', "unknown")))   
    audio_file.tags.add(TIT2(encoding=3, text=tags.get('title', "unknown"))) 


    track_cover = tags.get("cover", None)
    if track_cover is None:
        track_cover = tags.get("_albumart", None)


    if track_cover:
        if isinstance(track_cover, bytes):
            audio_file.tags.add(
                APIC(
                    encoding=3,
                    mime="image/jpeg",
                    type=3,
                    desc=u"cover",
                    data=track_cover
                )
                )
        elif isinstance(track_cover, str):
            try:
                async with aiohttp.ClientSession(headers=HTTP_HEADERS) as session:
                    async with session.get(track_cover) as response:
                        cover_data = await response.read()

                audio_file.tags.add(
                    APIC(
                        encoding=3,
                        mime="image/jpeg",
                        type=3,
                        desc=u"cover",
                        data=cover_data
                    )
                )
            except (aiohttp.ClientConnectorError, aiohttp.ClientConnectorSSLError, aiohttp.ServerTimeoutError):
                ...

    if isinstance(audio, io.BytesIO):
        return io.BytesIO(audio.getvalue())
    elif isinstance(audio, str):
        audio_file.save()
        return audio
    else:
        return '????'


from soundcloud import SoundCloud as Sound_Cloud
from soundcloud.resource.track import Track
from soundcloud.resource.user import User
from soundcloud.resource.playlist import AlbumPlaylist
from soundcloud import Media, Transcoding, Format, Badges, CreatorSubscription, Product, Visual, Visuals

import json, datetime, itertools, math, aiohttp, asyncio
from typing import Union

class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
        if isinstance(obj, Track):
            return obj.__dict__
        if isinstance(obj, Media):
            return obj.__dict__
        if isinstance(obj, Transcoding):
            return obj.__dict__
        if isinstance(obj, User):
            return obj.__dict__
        if isinstance(obj, Format):
            return obj.__dict__
        if isinstance(obj, Badges):
            return obj.__dict__
        if isinstance(obj, CreatorSubscription):
            return obj.__dict__
        if isinstance(obj, Product):
            return obj.__dict__
        if isinstance(obj, Visuals):
            return obj.__dict__
        if isinstance(obj, Visual):
            return obj.__dict__
        return super().default(obj)

class SoundCloud:
    def __init__(self, client_id:str, auth_token:str, debug:bool) -> None:
        self.debug = debug
        self.available_filters = ['track', 'user', 'album', 'playlist']
        self.methods = [
            'get_track',
            'download',
            'get_track_url'
        ]


        self.sc_client = Sound_Cloud(client_id = client_id, 
                                    auth_token = auth_token)

    async def __search_api_query(self, query:str,  filter:str, limit:int) -> Union[Track, User, AlbumPlaylist]:
        if filter == "track":
            data = self.sc_client.search_tracks(query = query)
        elif filter == "user":
            data = self.sc_client.search_users(query = query)
        elif filter == "album":
            data = self.sc_client.search_albums(query = query)
        elif filter == "playlist":
            data = self.sc_client.search_playlists(query = query)

        limited_data = itertools.islice(data, limit)
        return limited_data
    async def __api_query(self, method:str, **kwargs):
        data = None
        if method in self.methods:
            if method == 'get_track':
                data = self.sc_client.get_track(kwargs.get('track_id', 0))
            
            elif method == 'get_track_url':
                track_id = kwargs.get('track_id', 0)
                track = await self.get_track(track_id)
            
                download_url = None
                if track.downloadable:
                    download_url = self.sc_client.get_track_original_download(track.id, track.secret_token)


                if download_url is None:
                    aac_transcoding = None
                    mp3_transcoding = None
                    
                    for t in track.media.transcodings:
                        if t.format.protocol == "hls" and "aac" in t.preset:
                            aac_transcoding = t
                        elif t.format.protocol == "hls" and "mp3" in t.preset:
                            mp3_transcoding = t

                    transcoding = None

                    if aac_transcoding:
                        transcoding = aac_transcoding
                    elif mp3_transcoding:
                        transcoding = mp3_transcoding


                    if not transcoding:
                        return None
                    
                    url = transcoding.url
                    bitrate_KBps = 256 / 8 if "aac" in transcoding.preset else 128 / 8
                    total_bytes = bitrate_KBps * transcoding.duration
                    
                    min_size = 0
                    max_size = math.inf
                    
                    if not min_size <= total_bytes <= max_size:
                        return None
                    
                    if url is not None:
                        headers = self.sc_client.get_default_headers()
                        if self.sc_client.auth_token:
                            headers["Authorization"] = f"OAuth {self.sc_client.auth_token}"

                        async with aiohttp.ClientSession(headers=headers) as session:
                            async with session.get(url, params={"client_id": self.sc_client.client_id}) as response:
                                download_url = await response.json()   
                                download_url = download_url.get('url', "")

                return download_url
            
        return data
    async def __search_api_query(self, query:str,  filter:str, limit:int) -> Union[Track, User, AlbumPlaylist]:
        if filter == "track":
            data = self.sc_client.search_tracks(query = query)
        elif filter == "user":
            data = self.sc_client.search_users(query = query)
        elif filter == "album":
            data = self.sc_client.search_albums(query = query)
        elif filter == "playlist":
            data = self.sc_client.search_playlists(query = query)

        limited_data = itertools.islice(data, limit)
        return limited_data
    async def get_track_url(self, track_id:int, **kwargs) -> Track:
        """
            Getting information from a track by its ID

            :param track_id: Track ID on SoundCloud

            :return: SoundCloud.BasicTrack 
        """
        task = await asyncio.get_event_loop().run_in_executor(None, 
                    lambda: self.__api_query(
                                    method='get_track_url',
                                    track_id = track_id,**kwargs))
        
        
        task_result = await task
        return task_result
    async def get_track(self, track_id:int, **kwargs) -> Track:
        """
            Getting information from a track by its ID

            :param track_id: Track ID on SoundCloud

            :return: SoundCloud.BasicTrack 
        """
        task = await asyncio.get_event_loop().run_in_executor(None, 
                    lambda: self.__api_query(
                                    method='get_track',
                                    track_id = track_id,**kwargs))
        
        
        task_result = await task
        return task_result
    async def serialize_to_json(self, tracks:list):
        res = []
        for x in tracks:
            d = json.dumps(x, cls=CustomEncoder)
            res.append(json.loads(d))
        return res
    async def search(self, query:str, filter:str='track', limit:int=10, json_dump:bool=False):
        """
            Soundcloud search


            :param query: Search string
            :param filter: Filter for item types. Allowed values: ``track``, ``user``, ``album``, ``playlist``.
            Default: ``track``
            :param limit: Limit Results
            Default: ``10``

            :return: Generator limited by ``limit``
        """
        
        if filter not in self.available_filters:
            filter = self.available_filters[0]
        
        task = await asyncio.get_event_loop().run_in_executor(None, lambda: self.__search_api_query(query, filter, limit))
        task_result = await task
        if json_dump:
            task_result = await self.serialize_to_json(list(task_result))
        
        return task_result
    async def __search_by_link(self, link:str):
        return self.sc_client.resolve(link)
    async def search_by_link(self, link:str, json_dump:bool=False):
        task = await asyncio.get_event_loop().run_in_executor(None, lambda: self.__search_by_link(link))
        task_result = await task
        if json_dump:
            task_result = await self.serialize_to_json(list(task_result))
        
        return task_result



